# bash-ini: functions for handling ini files in bash
# Copyright (C)  2015-2024 Heinz Deinhart
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# bash-ini version 0.7.1

[ "$0" != "${BASH_SOURCE[0]}" ] || {
	echo "$0: this script must be sourced, not called" >&2
	exit 1
}

#@@ error messages

#@ error-codes (bash-ini)
# <50 .. reserved for base-base error-codes
#  51 .. illegal db-name
#  52 .. db named db-name not loaded
#  53 .. section not found
#  54 .. key not found
#  55 .. db named db-name already loaded
#  56 .. parse error
#  9x .. errors local to each function

#@ bini_string_of_error_code <ret-vn> <error-code>
# returns string representation of error code
# errors: 2, 3
function bini_error_string() {
    (( $# == 2 )) || {
	bb_ierr "bb_error_string: called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _err_code="$2"

    bb_vn_validate "$_ret_vn" || return 3

    if (( "$_err_code" < 50 ))
    then
	bb_error_string "$_ret_vn" "$_err_code"
    else
	local _msg

	case "$_err_code" in
	    51) _msg="illegal db-name";;
	    52) _msg="db named db-name not loaded";;
	    53) _msg="section not found";;
	    54) _msg="key not found";;
	    55) _msg="db named db-name already loaded";;
	    56) _msg="parse error";;
	    90|91|92|93|94|95|96|97|98|99) _msg="function local error";;
	    *) _msg="bb_error_string: unknown error code $_err_code";;
	esac
	printf -v "$_ret_vn" "%s" "$_msg"
    fi
}

# _bini_validate_db_name <caller-name> <db-name>
# errors: 1 (invalid db-name), 2
function _bini_validate_db_name() {
    (( $# == 1 )) || {
	bb_ierr "_bini_validate_db_name called with $# args instead of 1"
	return 2
    }

    [[ "$1" =~ ^[[:alnum:]_]+$ ]]
}

# _bini_validate_ret_vn <retvn>
# errors: 1 (invalid retvn), 2
function _bini_validate_ret_vn() {
    (( $# == 1 )) || {
	bb_ierr "_bini_validate_ret_vn called with $# args instead of 1"
	return 2
    }

    [[ "$1" =~ ^[[:alpha:]][[:alnum:]_]*$ ]]
}

#@@ accessing ini files

#@ bini_db_load <db-name> <ini-file>
# loads contents of ini-file
# db-name must match ^[[:alnum:]_]+$ and must be currently unset
# errors: 2, 8, 9, 51, 55, 56
function bini_db_load() {
    (( $# == 2 )) || {
	bb_ierr "bini_db_load called with $# args instead of 2"
	return 2
    }

    local _db_name="$1"
    local _inputfile="$2"

    _bini_validate_db_name "$_db_name" || {
	bb_msg debug "bini_db_load: abort: incorrect db-name: $_db_name"
	return 51
    }

    if bini_db_is_loaded "$_db_name"
    then
	bb_msg debug \
	    "bini_db_load: abort: db named $_db_name is already loaded/set"
	return 55
    fi

    [ -f "$_inputfile" ] || {
	bb_msg debug "bini_db_load: no such file: $_inputfile"
	return 8
    }
    [ -r "$_inputfile" ] || {
	bb_msg debug "bini_db_load: can not access file: $_inputfile"
	return 9
    }

    declare -g -A "bini_db_$_db_name"
    local -n _db="bini_db_$_db_name"

    local -i linenr=0
    local _section=

    local extglobsave
    extglobsave="$(shopt -p extglob || true)"
    shopt -s extglob

    while read -r line || [[ -n "$line" ]]
    do
	linenr=$((linenr + 1))

        # ignore empty lines and comments
	[[ "$line" =~ ^([[:space:]]*|[\;\#].*)$ ]] && {
	    continue
	}

	# sections
	[[ "$line" =~ ^\[([[:alnum:]_\-]([[:alnum:]\ _\-]*[[:alnum:]_\-])?)\][[:space:]]*$ ]] && {
	    _section="${BASH_REMATCH[1]}"
	    _db["${_section}§"]=''
	    continue
	}

	# values
	[[ "$line" =~ ^[[:space:]]*([[:alnum:]]([[:alnum:]\ _\-]*[[:alnum:]_\-])?)[[:blank:]]*=[[:blank:]]*(.*)[[:blank:]]*$ ]] && {
	    local _key="${BASH_REMATCH[1]}"
	    # shellcheck disable=SC2034
	    local _value="${BASH_REMATCH[3]}"
	    _db["${_section}§$_key"]="$_value"
	    continue
	}

	bb_msg debug "bini_db_load: parse error at line $linenr near \"$line\""
	eval "$extglobsave"
	return 56
    done < "$_inputfile"

    eval "$extglobsave"
}

#@ bini_db_destroy <db-name>
# destroys loaded ini db
# errors 2, 51
function bini_db_destroy() {
    (( $# == 1 )) || {
	bb_ierr "bini_db_destroy called with $# args instead of 1"
	return 2
    }

    local _db_name="$1"
    _bini_validate_db_name "$_db_name" || return 51
    unset "bini_db_$_db_name"
}

#@ bini_db_is_loaded <db-name>
# check if db is loaded
# errors 2, 51, 52
function bini_db_is_loaded() {
    (( $# == 1 )) || {
	bb_ierr "bini_db_is_loaded called with $# args instead of 1"
	return 2
    }

    local _db_name=$1
    _bini_validate_db_name "$_db_name" || return 51

    bb_dict_is_defined "bini_db_$_db_name" || return 52
}

#@@ data access

#@ bini_key_exists <db-name> <key> [<section>]
# check if a key exists
# if no section is specified only global context is searched
# errors: 2, 51, 52, 53, 54
function bini_key_exists() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "bini_key_exists needs 2 or 3 args but called with $#"
	return 2
    }

    local _db_name="$1"
    local _key="$2"
    local _section="${3-}"

    if [[ -z "$_section" ]]
    then
	bini_db_is_loaded "$_db_name" || return $? # returns 2, 51, 52
    else
	bini_section_exists "$_db_name" "$_section" || \
	    return $? # returns 2, 51, 52, 53
    fi

    bb_dict_contains_key "bini_db_$_db_name" "${_section}§$_key" || return 54
}

#@ bini_key_exists_with_default <db-name> <key> <section>
# like bini_key_exists but if section does exist and does not contain key it
# will search in global context
# errors: 2, 51, 52, 53, 54
function bini_key_exists_with_default() {
    (( $# == 3 )) || {
	bb_ierr "bini_key_exists_with_default needs 3 args but called with $#"
	return 2
    }

    local _db_name="$1"
    local _key="$2"
    local _section="$3"

    bini_section_exists "$_db_name" "$_section" || \
	return $? # returns 2, 51, 52, 53
    bini_key_exists "$_db_name" "$_key" "$_section" ||
	bini_key_exists "$_db_name" "$_key" # returns 2, 51, 52, 53, 54
}

#@ bini_section_exists <db-name> <section>
# check if a section exists
# errors: 2, 51, 52, 53
function bini_section_exists() {
    (( $# == 2 )) || {
	bb_ierr "bini_section_exists needs 2 args but called with $#"
	return 2
    }

    local _db_name="$1"
    local _section="$2"

    bini_db_is_loaded "$_db_name" || return $? # returns 2, 51, 52

    bb_dict_contains_key "bini_db_$_db_name" "${_section}§" || return 53
}

#@ bini_get_section_list <retval-name> <db-name>
# return a list of all sections as array into global array <retval-name>
# retval-name must match ^[[:alpha:]][[:alnum:]_]*$
# note: the order is undefined
# errors: 2, 3, 7, 51, 52
function bini_get_section_list() {
    (( $# == 2 )) || {
	bb_ierr "bini_get_section_list needs 2 args but called with $#"
	return 2
    }

    _bini_validate_ret_vn "$1" || return 3
    bb_array_is_defined "$1" || return 7
    bini_db_is_loaded "$2" || return $? # returns 2, 51, 52

    local -n _ret_ref="$1"
    # shellcheck disable=SC2178
    local -n _db="bini_db_$2"

    local -a _keys
    _keys=("${!_db[@]}")
    local -a _sections=()

    if (( ${#_keys[*]} > 0 )) # skip if file is empty
    then
	local _k
	for _k in "${_keys[@]}"
	do
	    [[ "$_k" =~ §$ ]] && _sections+=("${_k%§}")
	done
    fi

    _ret_ref=("${_sections[@]}")
}

#@ bini_get_key_list <retval-name> <db-name> [ <section> ]
# return a list of all keys as array into global array <retval-name>
# retval-name must match ^[[:alpha:]][[:alnum:]_]*$
# if no section is specified it will return keys found in global context
# note: the order is undefined
# errors: 2, 3, 51, 52, 53
function bini_get_key_list() {
    (( $# == 2 || $# == 3 )) || {
	bb_ierr "bini_get_key_list needs 3 or 4 args but called with $#"
	return 2
    }

    local _ret_vn="$1"
    local _db_name="$2"

    _bini_validate_ret_vn "$_ret_vn" || return 3
    bb_array_is_defined "$_ret_vn" || return 7

    # shellcheck disable=SC2178
    local -n _db="bini_db_$_db_name"

    local -a _keys=()
    local -a _dbkeys

    _dbkeys=("${!_db[@]}")

    if (( $# == 3 ))
    then
	# look in a section
	local _section="$3"

	bini_section_exists "$_db_name" "$_section" || \
	    return $? # returns 2, 51, 52, 53

	local _k
	for _k in "${_dbkeys[@]}"
	do
	    [[ "$_k" =~ ^${_section}§. ]] && _keys+=("${_k#*§}")
	done
    else
	# look in global context
	bini_db_is_loaded "$_db_name" || return $? # returns 2, 51, 52

	local _k
	for _k in "${_dbkeys[@]}"
	do
	    [[ "$_k" =~ ^§ ]] && _keys+=("${_k#§}")
	done
    fi

    # shellcheck disable=SC2178
    local -n _ret_ref="$_ret_vn"
    _ret_ref=("${_keys[@]}")
}

#@ bini_get_value <ret-vn> <db-name> <key> [ <section> ]
# return value for <key> in <db-name> in optional <section>
# note: if <section> is specified key must exist in it
# note: <retval-name> must not start with '_'
# errors: 2, 3, 51, 52, 53, 54
function bini_get_value() {
    (( $# == 3 || $# == 4 )) || {
	bb_ierr "bini_get_value needs 3 or 4 args but called with $#"
	return 2
    }

    local _ret_vn="$1"
    local _db_name="$2"
    local _key="$3"
    local _section="${4-}"

    _bini_validate_ret_vn "$_ret_vn" || return 3

    bini_key_exists "$_db_name" "$_key" "$_section" || \
	return $? # returns 2, 51, 52, 53, 54

    # shellcheck disable=SC2178
    local -n _db="bini_db_$_db_name"
    # shellcheck disable=SC2178
    local -n _ret_ref="$_ret_vn"

    # shellcheck disable=SC2178
    _ret_ref="${_db["${_section}§$_key"]}"
}

#@ bini_get_value_with_default <retval-name> <db-name> <key> <section>
# return value for <key> in <db-name> in <section>
# or try global if <section> exists but does not contain key
# note: <retval-name> must not start with '_'
# errors: 2, 3, 51, 52, 53, 54
function bini_get_value_with_default() {
    (( $# == 4 )) || {
	bb_ierr "bini_get_value_with_default needs 4 args but called with $#"
	return 2
    }

    local _ret_vn="$1"
    local _db_name="$2"
    local _key="$3"
    local _section="$4"

    _bini_validate_ret_vn "$_ret_vn" || return 3

    bini_get_value "$_ret_vn" "$_db_name" "$_key" "$_section" || {
	# returns 2, 3, 51, 52, 53, 54
	local _i="$?"
	case "$_i" in
	    51|52) return "$_i";;
	    53) # section was not found
		return 53;;
	    54) # section was found, but key not
		bini_get_value "$_ret_vn" "$_db_name" "$_key";;
	    *) # that should not happen
		bb_ierr "bini_get_value_with_default: got strange retcode: $_i"
		return 1;;
	esac
    }
}
